import functools
import sys
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from async_generator import asynccontextmanager
import trio
import trio_typing

from ._utils import get_task_name, iter_dag
from .abc import ManagerAPI, ServiceAPI
from .base import BaseManager
from .exceptions import DaemonTaskExit, LifecycleError, ServiceCancelled
from .typing import EXC_INFO


class TrioManager(BaseManager):
    # A nursery for sub tasks and services.  This nursery is cancelled if the
    # service is cancelled but allowed to exit normally if the service exits.
    _task_nursery: trio_typing.Nursery

    _service_task_dag: Dict[trio.hazmat.Task, List[trio.hazmat.Task]]
    _task_done_events: Dict[trio.hazmat.Task, trio.Event]
    _task_cancel_scopes: Dict[trio.hazmat.Task, trio.CancelScope]

    def __init__(self, service: ServiceAPI) -> None:
        super().__init__(service)

        # events
        self._started = trio.Event()
        self._cancelled = trio.Event()
        self._stopping = trio.Event()
        self._finished = trio.Event()

        # locks
        self._run_lock = trio.Lock()

        # DAG tracking
        self._service_task_dag = {}
        self._task_cancel_scopes = {}
        self._task_done_events = {}

    #
    # System Tasks
    #
    async def _handle_cancelled(self) -> None:
        self.logger.debug("%s: _handle_cancelled waiting for cancellation", self)
        await self._cancelled.wait()
        self.logger.debug("%s: _handle_cancelled triggering task cancellation", self)

        # Here we iterate over the task DAG such that we cancel the most deeply
        # nested tasks first ensuring that each has finished completely before
        # we move onto cancelling the parent tasks.
        #
        # We have to make a copy of the task dag because it is possible that
        # there is a task which has just been scheduled. In this case the new
        # task will end up being cancelled as part of it's parent task's cancel
        # scope, **or** if it was scheduled by an external API call it will be
        # cancelled as part of the global task nursery's cancellation.
        for task in iter_dag(self._service_task_dag.copy()):
            cancel_scope = self._task_cancel_scopes[task]
            done = self._task_done_events[task]
            cancel_scope.cancel()
            await done.wait()

        # This finaly cancellation of the task nursery's cancel scope ensures
        # that nothing is left behind and that the service will reliably exit.
        self._task_nursery.cancel_scope.cancel()

    async def _handle_run(self) -> None:
        """
        Run and monitor the actual :meth:`ServiceAPI.run` method.

        In the event that it throws an exception the service will be cancelled.
        """
        done, cancel_scope = self._track_current_task(parent=trio.hazmat.current_task())
        try:
            with cancel_scope:
                await self._service.run()
        except Exception as err:
            self.logger.debug(
                "%s: _handle_run got error, storing exception and setting cancelled: %s",
                self,
                err,
            )
            self._errors.append(cast(EXC_INFO, sys.exc_info()))
            self.cancel()
        else:
            # NOTE: Any service which uses daemon tasks will need to trigger
            # cancellation in order for the service to exit since this code
            # path does not trigger task cancellation.  It might make sense to
            # trigger cancellation if all of the running tasks are daemon
            # tasks.
            self.logger.debug(
                "%s: _handle_run exited cleanly, waiting for full stop...", self
            )
        finally:
            done.set()

    @classmethod
    async def run_service(cls, service: ServiceAPI) -> None:
        manager = cls(service)
        await manager.run()

    async def run(self) -> None:
        if self._run_lock.locked():
            raise LifecycleError(
                "Cannot run a service with the run lock already engaged.  Already started?"
            )
        elif self.is_started:
            raise LifecycleError("Cannot run a service which is already started.")

        try:
            async with self._run_lock:
                async with trio.open_nursery() as system_nursery:
                    system_nursery.start_soon(self._handle_cancelled)

                    try:
                        async with trio.open_nursery() as task_nursery:
                            self._task_nursery = task_nursery

                            task_nursery.start_soon(self._handle_run)

                            self._started.set()

                            # ***BLOCKING HERE***
                            # The code flow will block here until the background tasks have
                            # completed or cancellation occurs.
                    except Exception:
                        # Exceptions from any tasks spawned by our service will be caught by trio
                        # and raised here, so we store them to report together with any others we
                        # have already captured.
                        self._errors.append(cast(EXC_INFO, sys.exc_info()))
                    finally:
                        system_nursery.cancel_scope.cancel()

        finally:
            # We need this inside a finally because a trio.Cancelled exception may be raised
            # here and it wouldn't be swalled by the 'except Exception' above.
            self._stopping.set()
            self.logger.debug("%s stopping", self)
            self._finished.set()
            self.logger.debug("%s finished", self)

        # This is outside of the finally block above because we don't want to suppress
        # trio.Cancelled or trio.MultiError exceptions coming directly from trio.
        if self.did_error:
            raise trio.MultiError(
                tuple(
                    exc_value.with_traceback(exc_tb)
                    for _, exc_value, exc_tb in self._errors
                )
            )

    #
    # Event API mirror
    #
    @property
    def is_started(self) -> bool:
        return self._started.is_set()

    @property
    def is_cancelled(self) -> bool:
        return self._cancelled.is_set()

    @property
    def is_stopping(self) -> bool:
        return self._stopping.is_set() and not self.is_finished

    @property
    def is_finished(self) -> bool:
        return self._finished.is_set()

    #
    # Control API
    #
    def cancel(self) -> None:
        if not self.is_started:
            raise LifecycleError("Cannot cancel as service which was never started.")
        elif not self.is_running:
            return
        else:
            self._cancelled.set()

    #
    # Wait API
    #
    async def wait_started(self) -> None:
        await self._started.wait()

    async def wait_stopping(self) -> None:
        await self._stopping.wait()

    async def wait_finished(self) -> None:
        await self._finished.wait()

    def _track_current_task(
        self, parent: trio.hazmat.Task
    ) -> Tuple[trio.Event, trio.CancelScope]:
        current_task = trio.hazmat.current_task()
        if parent in self._service_task_dag:
            self._service_task_dag[parent].append(current_task)

        # We use an event to manually track when the child task is "done".
        # This is because trio has no API for awaiting completion of a task.
        done = trio.Event()
        self._task_done_events[current_task] = done

        # Each task gets its own `CancelScope` which is how we can manually
        # control cancellation order of the task DAG
        cancel_scope = trio.CancelScope()
        self._task_cancel_scopes[current_task] = cancel_scope

        # This data structure is setup so that child tasks can *inject*
        # themselves in as dependencies.
        self._service_task_dag[current_task] = []

        return done, cancel_scope

    async def _run_and_manage_task(
        self,
        async_fn: Callable[..., Awaitable[Any]],
        *args: Any,
        daemon: bool,
        name: str,
        parent: trio.hazmat.Task,
    ) -> None:
        self.logger.debug("running task '%s[daemon=%s]'", name, daemon)
        done, cancel_scope = self._track_current_task(parent)
        if parent not in self._service_task_dag:
            self.logger.debug("New root task %s[daemon=%s] added to DAG", name, daemon)

        try:
            with cancel_scope:
                try:
                    await async_fn(*args)
                except Exception as err:
                    self.logger.debug(
                        "task '%s[daemon=%s]' exited with error: %s",
                        name,
                        daemon,
                        err,
                        exc_info=True,
                    )
                    self._errors.append(cast(EXC_INFO, sys.exc_info()))
                    self.cancel()
                else:
                    self.logger.debug("task '%s[daemon=%s]' finished.", name, daemon)
                    if daemon:
                        self.logger.debug(
                            "daemon task '%s' exited unexpectedly.  Cancelling service: %s",
                            name,
                            self,
                        )
                        self.cancel()
                        raise DaemonTaskExit(f"Daemon task {name} exited")
        finally:
            done.set()

    def run_task(
        self,
        async_fn: Callable[..., Awaitable[Any]],
        *args: Any,
        daemon: bool = False,
        name: str = None,
    ) -> None:
        if not self.is_running:
            raise LifecycleError(
                "Tasks may not be scheduled if the service is not running"
            )

        task_name = get_task_name(async_fn, name)
        if self.is_running and self.is_cancelled:
            self.logger.debug(
                "Service is in the process of being cancelled. Not running task "
                "%s[daemon=%s]",
                task_name,
                daemon,
            )
            return

        current_task = trio.hazmat.current_task()

        self._task_nursery.start_soon(
            functools.partial(
                self._run_and_manage_task,
                daemon=daemon,
                name=task_name,
                parent=current_task,
            ),
            async_fn,
            *args,
            name=task_name,
        )

    def run_child_service(
        self, service: ServiceAPI, daemon: bool = False, name: str = None
    ) -> ManagerAPI:
        child_manager = type(self)(service)
        task_name = get_task_name(service, name)
        self.run_task(child_manager.run, daemon=daemon, name=task_name)
        return child_manager


TFunc = TypeVar("TFunc", bound=Callable[..., Coroutine[Any, Any, Any]])


_ChannelPayload = Tuple[Optional[Any], Optional[BaseException]]


async def _wait_stopping_or_finished(
    service: ServiceAPI,
    api_func: Callable[..., Any],
    channel: trio.abc.SendChannel[_ChannelPayload],
) -> None:
    manager = service.manager

    if service.manager.is_stopping or service.manager.is_finished:
        await channel.send(
            (
                None,
                ServiceCancelled(
                    f"Cannot access external API {api_func}.  Service is not running: "
                    f"started={manager.is_started}  running={manager.is_running} "
                    f"stopping={manager.is_stopping}  finished={manager.is_finished}"
                ),
            )
        )
        return

    await service.manager.wait_stopping()
    await channel.send(
        (
            None,
            ServiceCancelled(
                f"Cannot access external API {api_func}.  Service is not running: "
                f"started={manager.is_started}  running={manager.is_running} "
                f"stopping={manager.is_stopping}  finished={manager.is_finished}"
            ),
        )
    )


async def _wait_api_fn(
    self: ServiceAPI,
    api_fn: Callable[..., Any],
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
    channel: trio.abc.SendChannel[_ChannelPayload],
) -> None:
    try:
        result = await api_fn(self, *args, **kwargs)
    except Exception:
        _, exc_value, exc_tb = sys.exc_info()
        if exc_value is None or exc_tb is None:
            raise Exception(
                "This should be unreachable but acts as a type guard for mypy"
            )
        await channel.send((None, exc_value.with_traceback(exc_tb)))
    else:
        await channel.send((result, None))


def external_api(func: TFunc) -> TFunc:
    @functools.wraps(func)
    async def inner(self: ServiceAPI, *args: Any, **kwargs: Any) -> Any:
        if not hasattr(self, "manager"):
            raise ServiceCancelled(
                f"Cannot access external API {func}.  Service has not been run."
            )

        manager = self.manager

        if not manager.is_running:
            raise ServiceCancelled(
                f"Cannot access external API {func}.  Service is not running: "
                f"started={manager.is_started}  running={manager.is_running} "
                f"stopping={manager.is_stopping}  finished={manager.is_finished}"
            )

        channels: Tuple[
            trio.abc.SendChannel[_ChannelPayload],
            trio.abc.ReceiveChannel[_ChannelPayload],
        ] = trio.open_memory_channel(0)
        send_channel, receive_channel = channels

        async with trio.open_nursery() as nursery:
            # mypy's type hints for start_soon break with this invocation.
            nursery.start_soon(
                _wait_api_fn, self, func, args, kwargs, send_channel  # type: ignore
            )
            nursery.start_soon(_wait_stopping_or_finished, self, func, send_channel)
            result, err = await receive_channel.receive()
            nursery.cancel_scope.cancel()
        if err is None:
            return result
        else:
            raise err

    return cast(TFunc, inner)


@asynccontextmanager
async def background_trio_service(service: ServiceAPI) -> AsyncIterator[ManagerAPI]:
    """
    Run a service in the background.

    The service is running within the context
    block and will be properly cleaned up upon exiting the context block.
    """
    async with trio.open_nursery() as nursery:
        manager = TrioManager(service)
        nursery.start_soon(manager.run)
        await manager.wait_started()
        try:
            yield manager
        finally:
            await manager.stop()



from .abc import ManagerAPI, ServiceAPI  # noqa: F401
from .asyncio import (  # noqa: F401
    AsyncioManager,
    background_asyncio_service,
    external_api as external_asyncio_api,
)
from .base import Service, as_service  # noqa: F401
from .exceptions import DaemonTaskExit, LifecycleError, ServiceCancelled  # noqa: F401
from .trio import TrioManager, background_trio_service  # noqa: F401

run_asyncio_service = AsyncioManager.run_service
run_trio_service = TrioManager.run_service



import asyncio
import sys

if sys.version_info >= (3, 7):
    # This API isn't available in 3.6
    get_current_task = asyncio.current_task
else:
    # This API is deprecated in 3.7+
    get_current_task = asyncio.Task.current_task



from abc import abstractmethod
import asyncio
import logging
import random
from typing import Any, Dict, Tuple, Type, Union

import trio

from async_service import Service

DAG = Dict[int, Tuple[int, ...]]

DEFAULT_DAG: DAG = {
    0: (1, 2, 3),
    1: (),
    2: (4, 5),
    3: (),
    4: (6, 7, 8, 9, 10),
    5: (),
    6: (),
    7: (),
    8: (11,),
    9: (),
    10: (),
    11: (12,),
    12: (13,),
    13: (),
}


class Resource:
    is_active = None
    was_checked = False

    async def __aenter__(self) -> "Resource":
        self.is_active = True
        return self

    async def __aexit__(self, *args: Any) -> None:
        self.is_active = False


class DAGServiceTest(Service):
    """
    This service is for testing whether the task DAG gets shutdown in the
    correct order.
    """

    all_checks_passed = False

    logger = logging.getLogger("async_service.testing.DAGServiceTest")

    event_class: Union[Type[trio.Event], Type[asyncio.Event]]

    @abstractmethod
    async def yield_execution(self, count: int) -> None:
        ...

    @abstractmethod
    async def ready_cancel(self) -> None:
        ...

    def __init__(self, dag: DAG = None):
        if dag is None:
            self._dag = DEFAULT_DAG
        else:
            self._dag = dag

        # A set of tasks that signal that all of the dag child tasks are at a
        # state in which they can be cancelled.
        self._child_tasks_all_ready_events = {
            task_id: self.event_class() for task_id in self._dag.keys()
        }
        self._task_resources = {task_id: Resource() for task_id in self._dag.keys()}

    async def _do_task(self, task_id: int, resource: Resource) -> None:
        self.logger.info("task-%d: START", task_id)
        assert not resource.is_active
        assert not resource.was_checked
        try:
            async with resource:
                assert resource.is_active
                # run the children of this task
                self.manager.run_task(
                    self._do_child, task_id, resource, name=f"child-{task_id}"
                )
                await self.manager.wait_finished()
        finally:
            self.logger.info("task-%d: EXITING", task_id)
            # we yield execution a random number of times to ensure that the
            # order that these tasks were cued doesn't effect the desired
            # functionality.
            await self.yield_execution(random.randint(0, 100))
            assert not resource.is_active
            assert resource.was_checked
            self.logger.info("task-%d: FINISH", task_id)

    async def _do_child(self, task_id: int, resource: Resource) -> None:
        self.logger.info("child--%d: START", task_id)
        assert resource.is_active
        assert not resource.was_checked

        try:
            await self._run_children(task_id)
            self._child_tasks_all_ready_events[task_id].set()
            self.logger.info("child-%d: RUNNING", task_id)
            await self.manager.wait_finished()
        finally:
            self.logger.info("child-%d: EXITING", task_id)
            # we yield execution a random number of times to ensure that the
            # order that these tasks were cued doesn't effect the desired
            # functionality.
            await self.yield_execution(random.randint(0, 100))
            assert resource.is_active
            resource.was_checked = True
            self.logger.info("child-%d: FINISHED", task_id)

    async def _run_children(self, task_id: int) -> None:
        child_tasks = self._dag[task_id]
        for child_id in child_tasks:
            resource = self._task_resources[child_id]
            self.manager.run_task(
                self._do_task, child_id, resource, name=f"task-{task_id}"
            )
            is_running_event = self._child_tasks_all_ready_events[child_id]
            await is_running_event.wait()

    async def run(self) -> None:
        # pre-run sanity checks
        assert all(
            resource.is_active is None for resource in self._task_resources.values()
        )
        assert all(
            resource.was_checked is False for resource in self._task_resources.values()
        )
        resource = self._task_resources[0]
        try:
            await self._do_task(0, resource)
        finally:
            # post-run sanity checks
            for task_id, resource in self._task_resources.items():
                if resource.is_active is not False:
                    raise AssertionError(
                        f"Resource for task-{task_id} was not `False`: {resource.is_active!r}"
                    )
            for task_id, resource in self._task_resources.items():
                if resource.was_checked is not True:
                    raise AssertionError(f"Resource for task-{task_id} was not checked")
            self.all_checks_passed = True



import asyncio
import functools
import sys
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Set,
    TypeVar,
    cast,
)

from async_generator import asynccontextmanager
from trio import MultiError

from ._utils import get_task_name, iter_dag
from .abc import ManagerAPI, ServiceAPI
from .asyncio_compat import get_current_task
from .base import BaseManager
from .exceptions import DaemonTaskExit, LifecycleError, ServiceCancelled
from .typing import EXC_INFO


class AsyncioManager(BaseManager):
    # Tracking of the system level background tasks.
    _system_tasks: Set["asyncio.Future[None]"]

    # Tracking of the background tasks that the service has initiated.
    _service_task_dag: Dict["asyncio.Future[Any]", List["asyncio.Future[Any]"]]

    def __init__(
        self, service: ServiceAPI, loop: asyncio.AbstractEventLoop = None
    ) -> None:
        super().__init__(service)

        self._loop = loop

        # events
        self._started = asyncio.Event()
        self._cancelled = asyncio.Event()
        self._stopping = asyncio.Event()
        self._finished = asyncio.Event()

        # locks
        self._run_lock = asyncio.Lock()

        # task tracking
        self._service_task_dag = {}
        self._system_tasks = set()

    #
    # System Tasks
    #
    async def _handle_cancelled(self) -> None:
        """
        When cancellation is requested this triggers the cancellation of all
        background tasks.
        """
        self.logger.debug("%s: _handle_cancelled waiting for cancellation", self)
        await self._cancelled.wait()
        self.logger.debug("%s: _handle_cancelled triggering task cancellation", self)

        # Here we iterate over the task DAG such that we cancel the most deeply
        # nested tasks first ensuring that each has finished completely before
        # we move onto cancelling the parent tasks.
        #
        # We have to make a copy of the task dag because it is possible that
        # there is a task which has just been scheduled. In this case the new
        # task will end up being cancelled as part of it's parent task's cancel
        # scope, **or** if it was scheduled by an external API call it will be
        # cancelled as part of the global task nursery's cancellation.
        for task in iter_dag(self._service_task_dag.copy()):
            if not task.done():
                task.cancel()

            try:
                await task
            except asyncio.CancelledError:
                pass
            except Exception:
                self._errors.append(cast(EXC_INFO, sys.exc_info()))

    async def _handle_stopping(self) -> None:
        """
        Once the `_stopping` event is set this triggers cancellation of the system tasks.
        """
        self.logger.debug("%s: _handle_stopping waiting for stopping", self)
        await self.wait_stopping()
        self.logger.debug(
            "%s: _handle_stopping triggering system task cancellations", self
        )

        # trigger cancellation of all of the system tasks
        for task in self._system_tasks:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
            except Exception:
                self._errors.append(cast(EXC_INFO, sys.exc_info()))

    @classmethod
    async def run_service(
        cls, service: ServiceAPI, loop: asyncio.AbstractEventLoop = None
    ) -> None:
        manager = cls(service, loop=loop)
        await manager.run()

    async def run(self) -> None:
        if self._run_lock.locked():
            raise LifecycleError(
                "Cannot run a service with the run lock already engaged.  Already started?"
            )
        elif self.is_started:
            raise LifecycleError("Cannot run a service which is already started.")

        async with self._run_lock:
            try:
                handle_cancelled_task = asyncio.ensure_future(
                    self._handle_cancelled(), loop=self._loop
                )
                handle_stopping_task = asyncio.ensure_future(
                    self._handle_stopping(), loop=self._loop
                )

                self._system_tasks.add(handle_cancelled_task)
                self._system_tasks.add(handle_stopping_task)

                self._started.set()

                self.run_task(self._service.run)

                # block here until all service tasks have finished
                await self._wait_service_tasks()
            finally:
                # None of the statements above are supposed to raise exceptions (as they all rely
                # on _run_and_manage_task() to swallow and store them in self._errors, but in case
                # they do we probably have a bug so we just mark the service as stopping and
                # return.
                self._stopping.set()
                self.logger.debug("%s stopping", self)

            # block here until all system tasks have finished.
            await asyncio.wait(
                (handle_cancelled_task, handle_stopping_task),
                return_when=asyncio.ALL_COMPLETED,
            )

        self._finished.set()
        self.logger.debug("%s finished", self)

        # Above we rely on run_task() and handle_cancelled()/handle_stopping() to run the
        # service/tasks and swallow/collect exceptions so that they can be reported all together
        # here.
        if self.did_error:
            raise MultiError(
                tuple(
                    exc_value.with_traceback(exc_tb)
                    for _, exc_value, exc_tb in self._errors
                )
            )

    async def _wait_service_tasks(self) -> None:
        while True:
            done, pending = await asyncio.wait(
                tuple(self._service_task_dag.keys()), return_when=asyncio.ALL_COMPLETED
            )
            if all(task.done() for task in self._service_task_dag.keys()):
                break

    #
    # Event API mirror
    #
    @property
    def is_started(self) -> bool:
        return self._started.is_set()

    @property
    def is_cancelled(self) -> bool:
        return self._cancelled.is_set()

    @property
    def is_stopping(self) -> bool:
        return self._stopping.is_set() and not self.is_finished

    @property
    def is_finished(self) -> bool:
        return self._finished.is_set()

    #
    # Control API
    #
    def cancel(self) -> None:
        if not self.is_started:
            raise LifecycleError("Cannot cancel as service which was never started.")
        self._cancelled.set()

    #
    # Wait API
    #
    async def wait_started(self) -> None:
        await self._started.wait()

    async def wait_stopping(self) -> None:
        await self._stopping.wait()

    async def wait_finished(self) -> None:
        await self._finished.wait()

    async def _run_and_manage_task(
        self,
        async_fn: Callable[..., Awaitable[Any]],
        *args: Any,
        daemon: bool,
        name: str,
    ) -> None:
        self.logger.debug("running task '%s[daemon=%s]'", name, daemon)
        try:
            await async_fn(*args)
        except asyncio.CancelledError:
            raise
        except Exception as err:
            self.logger.debug(
                "task '%s[daemon=%s]' exited with error: %s",
                name,
                daemon,
                err,
                exc_info=True,
            )
            self._errors.append(cast(EXC_INFO, sys.exc_info()))
            self.cancel()
        else:
            self.logger.debug("task '%s[daemon=%s]' finished.", name, daemon)
            if daemon:
                self.logger.debug(
                    "daemon task '%s' exited unexpectedly.  Cancelling service: %s",
                    name,
                    self,
                )
                self.cancel()
                raise DaemonTaskExit(f"Daemon task {name} exited")

    def run_task(
        self,
        async_fn: Callable[..., Awaitable[Any]],
        *args: Any,
        daemon: bool = False,
        name: str = None,
    ) -> None:
        if not self.is_running:
            raise LifecycleError(
                "Tasks may not be scheduled if the service is not running"
            )

        task_name = get_task_name(async_fn, name)

        if self.is_running and self.is_cancelled:
            self.logger.debug(
                "Service is in the process of being cancelled. Not running task "
                "%s[daemon=%s]",
                task_name,
                daemon,
            )
            return

        task = asyncio.ensure_future(
            self._run_and_manage_task(async_fn, *args, daemon=daemon, name=task_name),
            loop=self._loop,
        )
        self._service_task_dag[task] = []

        parent_task = get_current_task()
        if parent_task in self._service_task_dag:
            self._service_task_dag[parent_task].append(task)
        else:
            self.logger.debug(
                "New root task %s[daemon=%s] added to DAG", task_name, daemon
            )

    def run_child_service(
        self, service: ServiceAPI, daemon: bool = False, name: str = None
    ) -> ManagerAPI:
        child_manager = type(self)(service, loop=self._loop)
        task_name = get_task_name(service, name)
        self.run_task(child_manager.run, daemon=daemon, name=task_name)
        return child_manager


@asynccontextmanager
async def cleanup_tasks(*tasks: "asyncio.Future[Any]") -> AsyncIterator[None]:
    """
    Context manager that ensures that all tasks are properly cancelled and awaited.

    The order in which tasks are cleaned is such that the first task will be
    the last to be cancelled/awaited.

    This function **must** be called with at least one task.
    """
    try:
        task = tasks[0]
    except IndexError:
        raise TypeError("cleanup_tasks must be called with at least one task")

    try:
        if len(tasks) > 1:
            async with cleanup_tasks(*tasks[1:]):
                yield
        else:
            yield
    finally:
        if not task.done():
            task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass


TFunc = TypeVar("TFunc", bound=Callable[..., Coroutine[Any, Any, Any]])


def external_api(func: TFunc) -> TFunc:
    @functools.wraps(func)
    async def inner(self: ServiceAPI, *args: Any, **kwargs: Any) -> Any:
        if not hasattr(self, "manager"):
            raise ServiceCancelled(
                f"Cannot access external API {func}.  Service has not been run."
            )

        manager = self.manager

        if not manager.is_running:
            raise ServiceCancelled(
                f"Cannot access external API {func}.  Service is not running: "
                f"started={manager.is_started}  running={manager.is_running} "
                f"stopping={manager.is_stopping}  finished={manager.is_finished}"
            )

        func_task: "asyncio.Future[Any]" = asyncio.ensure_future(
            func(self, *args, **kwargs)
        )
        service_stopping_task = asyncio.ensure_future(manager.wait_stopping())

        done, pending = await asyncio.wait(
            (func_task, service_stopping_task), return_when=asyncio.FIRST_COMPLETED
        )
        async with cleanup_tasks(*done, *pending):
            if func_task.done():
                return await func_task
            elif service_stopping_task.done():
                raise ServiceCancelled(
                    f"Cannot access external API {func}.  Service is not running: "
                    f"started={manager.is_started}  running={manager.is_running} "
                    f"stopping={manager.is_stopping}  finished={manager.is_finished}"
                )
            else:
                raise Exception("Code path should be unreachable")

    return cast(TFunc, inner)


@asynccontextmanager
async def background_asyncio_service(
    service: ServiceAPI, loop: asyncio.AbstractEventLoop = None
) -> AsyncIterator[ManagerAPI]:
    """
    Run a service in the background.

    The service is running within the context block and will be properly
    cleaned up upon exiting the context block.
    """
    manager = AsyncioManager(service, loop=loop)
    task = asyncio.ensure_future(manager.run(), loop=loop)

    try:
        async with cleanup_tasks(task):
            await manager.wait_started()

            try:
                yield manager
            finally:
                await manager.stop()
    finally:
        if manager.did_error:
            # TODO: better place for this.
            raise MultiError(
                tuple(
                    exc_value.with_traceback(exc_tb)
                    for _, exc_value, exc_tb in manager._errors
                )
            )
